"""initial schema

Revision ID: 48be4072fe58
Revises:
Create Date: 2022-06-01 18:37:12.396138

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "48be4072fe58"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "conda_channel",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("name", sa.Unicode(length=255), nullable=False, unique=True),
        sa.Column("last_update", sa.DateTime(), nullable=True),
    )

    op.create_table(
        "conda_store_configuration",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("disk_usage", sa.BigInteger(), nullable=True),
        sa.Column("free_storage", sa.BigInteger(), nullable=True),
        sa.Column("total_storage", sa.BigInteger(), nullable=True),
    )

    op.create_table(
        "namespace",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("name", sa.Unicode(length=255), nullable=True, unique=True),
        sa.Column("deleted_on", sa.DateTime(), nullable=True),
    )

    op.create_table(
        "specification",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("name", sa.Unicode(length=255), nullable=False),
        sa.Column("spec", sa.JSON(), nullable=False),
        sa.Column("sha256", sa.Unicode(length=255), nullable=False, unique=True),
        sa.Column("created_on", sa.DateTime(), nullable=True),
    )

    op.create_table(
        "conda_package",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("channel_id", sa.Integer(), nullable=True),
        sa.Column("build", sa.Unicode(length=64), nullable=False),
        sa.Column("build_number", sa.Integer(), nullable=False),
        sa.Column("constrains", sa.JSON(), nullable=True),
        sa.Column("depends", sa.JSON(), nullable=False),
        sa.Column("license", sa.Text(), nullable=True),
        sa.Column("license_family", sa.Unicode(length=64), nullable=True),
        sa.Column("md5", sa.Unicode(length=255), nullable=False),
        sa.Column("name", sa.Unicode(length=255), nullable=False),
        sa.Column("sha256", sa.Unicode(length=64), nullable=False),
        sa.Column("size", sa.BigInteger(), nullable=False),
        sa.Column("subdir", sa.Unicode(length=64), nullable=True),
        sa.Column("timestamp", sa.BigInteger(), nullable=True),
        sa.Column("version", sa.Unicode(length=64), nullable=False),
        sa.Column("summary", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
    )

    op.create_table(
        "environment",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("namespace_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.Unicode(length=255), nullable=False),
        sa.Column("current_build_id", sa.Integer(), nullable=True),
        sa.Column("deleted_on", sa.DateTime(), nullable=True),
    )

    op.create_table(
        "solve",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("specification_id", sa.Integer(), nullable=False),
        sa.Column("scheduled_on", sa.DateTime(), nullable=True),
        sa.Column("started_on", sa.DateTime(), nullable=True),
        sa.Column("ended_on", sa.DateTime(), nullable=True),
    )

    op.create_table(
        "build",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("specification_id", sa.Integer(), nullable=False),
        sa.Column("environment_id", sa.Integer(), nullable=False),
        sa.Column(
            "status",
            sa.Enum("QUEUED", "BUILDING", "COMPLETED", "FAILED", name="buildstatus"),
            nullable=True,
        ),
        sa.Column("size", sa.BigInteger(), nullable=True),
        sa.Column("scheduled_on", sa.DateTime(), nullable=True),
        sa.Column("started_on", sa.DateTime(), nullable=True),
        sa.Column("ended_on", sa.DateTime(), nullable=True),
        sa.Column("deleted_on", sa.DateTime(), nullable=True),
    )

    op.create_table(
        "solve_conda_package",
        sa.Column("solve_id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("conda_package_id", sa.Integer(), nullable=False, primary_key=True),
    )

    op.create_table(
        "build_artifact",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("build_id", sa.Integer(), nullable=True),
        sa.Column(
            "artifact_type",
            sa.Enum(
                "DIRECTORY",
                "LOCKFILE",
                "LOGS",
                "YAML",
                "CONDA_PACK",
                "DOCKER_BLOB",
                "DOCKER_MANIFEST",
                name="buildartifacttype",
            ),
            nullable=False,
        ),
        sa.Column("key", sa.Unicode(length=255), nullable=True),
    )

    op.create_table(
        "build_conda_package",
        sa.Column("build_id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("conda_package_id", sa.Integer(), nullable=False, primary_key=True),
    )

    with op.batch_alter_table(
        "conda_package",
        table_args=[
            sa.ForeignKeyConstraint(
                ["channel_id"],
                ["conda_channel.id"],
            ),
            sa.UniqueConstraint(
                "channel_id",
                "subdir",
                "name",
                "version",
                "build",
                "build_number",
                "sha256",
                name="_conda_package_uc",
            ),
        ],
    ):
        pass

    with op.batch_alter_table(
        "environment",
        table_args=[
            sa.ForeignKeyConstraint(["current_build_id"], ["build.id"]),
            sa.UniqueConstraint("namespace_id", "name", name="_namespace_name_uc"),
            sa.ForeignKeyConstraint(["namespace_id"], ["namespace.id"]),
        ],
    ):
        pass

    with op.batch_alter_table(
        "solve",
        table_args=[
            sa.ForeignKeyConstraint(
                ["specification_id"],
                ["specification.id"],
            ),
        ],
    ):
        pass

    with op.batch_alter_table(
        "build",
        table_args=[
            sa.ForeignKeyConstraint(["environment_id"], ["environment.id"]),
            sa.ForeignKeyConstraint(["specification_id"], ["specification.id"]),
        ],
    ):
        pass

    with op.batch_alter_table(
        "solve_conda_package",
        table_args=[
            sa.ForeignKeyConstraint(
                ["conda_package_id"], ["conda_package.id"], ondelete="CASCADE"
            ),
            sa.ForeignKeyConstraint(["solve_id"], ["solve.id"], ondelete="CASCADE"),
        ],
    ):
        pass

    with op.batch_alter_table(
        "build_artifact",
        table_args=[
            sa.ForeignKeyConstraint(
                ["build_id"],
                ["build.id"],
            ),
        ],
    ):
        pass

    with op.batch_alter_table(
        "build_conda_package",
        table_args=[
            sa.ForeignKeyConstraint(["build_id"], ["build.id"], ondelete="CASCADE"),
            sa.ForeignKeyConstraint(
                ["conda_package_id"], ["conda_package.id"], ondelete="CASCADE"
            ),
        ],
    ):
        pass


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("build_conda_package")
    op.drop_table("build_artifact")
    op.drop_table("solve_conda_package")
    op.drop_table("build")
    op.drop_table("solve")
    op.drop_table("environment")
    op.drop_table("conda_package")
    op.drop_table("specification")
    op.drop_table("namespace")
    op.drop_table("conda_store_configuration")
    op.drop_table("conda_channel")
    # ### end Alembic commands ###



import datetime
import os
import stat
import subprocess
import pathlib
import tempfile
import traceback

import filelock
import yaml

from typing import Dict, Union

from conda_store_server import api, conda, orm, utils, schema


def set_build_started(conda_store, build):
    build.status = schema.BuildStatus.BUILDING
    build.started_on = datetime.datetime.utcnow()
    conda_store.db.commit()


def set_build_failed(conda_store, build, logs):
    conda_store.storage.set(
        conda_store.db,
        build.id,
        build.log_key,
        logs,
        content_type="text/plain",
        artifact_type=schema.BuildArtifactType.LOGS,
    )
    build.status = schema.BuildStatus.FAILED
    build.ended_on = datetime.datetime.utcnow()
    conda_store.db.commit()


def set_build_completed(conda_store, build, logs, packages):
    package_keys = [
        "channel_id",
        "license",
        "license_family",
        "name",
        "version",
        "summary",
        "description",
    ]
    package_build_keys = [
        "build",
        "build_number",
        "constrains",
        "depends",
        "md5",
        "sha256",
        "size",
        "subdir",
        "timestamp",
    ]

    for package in packages:

        channel = package["channel_id"]
        if channel == "https://conda.anaconda.org/pypi":
            # ignore pypi package for now
            continue

        channel_orm = api.get_conda_channel(conda_store.db, channel)
        if channel_orm is None:
            # Empty list for conda_allowed_channels allows any channel ( PR #358 )
            if len(conda_store.conda_allowed_channels) == 0:
                channel_orm = api.create_conda_channel(conda_store.db, channel)
                conda_store.db.commit()
            else:
                raise ValueError(
                    f"channel url={channel} not recognized in conda-store channel database"
                )
        package["channel_id"] = channel_orm.id

        # Retrieve the package from the DB if it already exists
        _package = (
            conda_store.db.query(orm.CondaPackage)
            .filter(orm.CondaPackage.channel_id == package["channel_id"])
            .filter(orm.CondaPackage.name == package["name"])
            .filter(orm.CondaPackage.version == package["version"])
            .first()
        )

        # If it doesn't exist, let's create it in DB
        if _package is None:
            package_dict = {k: package[k] for k in package_keys}
            _package = orm.CondaPackage(**package_dict)
            conda_store.db.add(_package)

        # Retrieve the build for this pacakge, if it already exists
        _package_build = (
            conda_store.db.query(orm.CondaPackageBuild)
            .filter(orm.CondaPackageBuild.package == _package)
            .filter(orm.CondaPackageBuild.md5 == package["md5"])
            .first()
        )

        # If it doesn't exist, let's create it in DB
        if _package_build is None:
            package_build_dict = {k: package[k] for k in package_build_keys}
            # Attach the package_build to its package
            package_build_dict["package_id"] = _package.id
            _package_build = orm.CondaPackageBuild(**package_build_dict)
            conda_store.db.add(_package_build)

        build.package_builds.append(_package_build)
        conda_store.db.commit()

    conda_store.storage.set(
        conda_store.db,
        build.id,
        build.log_key,
        logs,
        content_type="text/plain",
        artifact_type=schema.BuildArtifactType.LOGS,
    )
    build.status = schema.BuildStatus.COMPLETED
    build.ended_on = datetime.datetime.utcnow()

    # add records for lockfile and directory build artifacts
    lockfile_build_artifact = orm.BuildArtifact(
        build_id=build.id, artifact_type=schema.BuildArtifactType.LOCKFILE, key=""
    )
    directory_build_artifact = orm.BuildArtifact(
        build_id=build.id,
        artifact_type=schema.BuildArtifactType.DIRECTORY,
        key=build.build_path(conda_store),
    )
    conda_store.db.add(lockfile_build_artifact)
    conda_store.db.add(directory_build_artifact)

    environment = (
        conda_store.db.query(orm.Environment)
        .filter(orm.Environment.name == build.specification.name)
        .first()
    )
    environment.current_build = build
    environment.specification = build.specification
    conda_store.db.commit()


def build_environment(conda_command, environment_filename, conda_prefix):

    return subprocess.check_output(
        [
            conda_command,
            "env",
            "create",
            "-p",
            conda_prefix,
            "-f",
            str(environment_filename),
        ],
        stderr=subprocess.STDOUT,
        encoding="utf-8",
    )


def set_conda_environment_variables(
    conda_prefix: pathlib.Path, environment_variables: Dict[str, Union[str, int]]
):
    """Takes an input of the conda prefix and the, variables defined in the environment yaml
    specification. Then, generates the files neccesary to "activate" these when an environment
    is activated.
    """
    for item in ("activate", "deactivate"):
        folderpath = conda_prefix.joinpath("etc", "conda", f"{item}.d")
        folderpath.mkdir(parents=True, exist_ok=False)
        env_vars_file = folderpath.joinpath("env_vars.sh")
        env_vars_file.touch()
        with open(env_vars_file, "w") as f:
            f.write("#!/bin/bash\n")
            if item == "activate":
                for key in environment_variables:
                    f.write(f"export {key}={environment_variables[key]}\n")
            elif item == "deactivate":
                for key in environment_variables.keys():
                    f.write(f"unset {key}\n")
    return


def build_conda_environment(conda_store, build):
    """Build a conda environment with set uid/gid/and permissions and
    symlink the build to a named environment

    """
    set_build_started(conda_store, build)

    conda_prefix = build.build_path(conda_store)
    os.makedirs(os.path.dirname(conda_prefix), exist_ok=True)

    environment_prefix = build.environment_path(conda_store)
    os.makedirs(os.path.dirname(environment_prefix), exist_ok=True)

    conda_store.log.info(f"building conda environment={conda_prefix}")

    try:
        with utils.timer(conda_store.log, f"building {conda_prefix}"):
            with tempfile.TemporaryDirectory() as tmpdir:
                tmp_environment_filename = os.path.join(tmpdir, "environment.yaml")
                with open(tmp_environment_filename, "w") as f:
                    yaml.dump(build.specification.spec, f)
                    if conda_store.serialize_builds:
                        with filelock.FileLock(
                            os.path.join(tempfile.tempdir, "conda-store.lock")
                        ):
                            output = build_environment(
                                conda_store.conda_command,
                                tmp_environment_filename,
                                conda_prefix,
                            )
                    else:
                        output = build_environment(
                            conda_store.conda_command,
                            tmp_environment_filename,
                            conda_prefix,
                        )

                    if build.specification.spec.get("variables") is not None:
                        set_conda_environment_variables(
                            pathlib.Path(conda_prefix),
                            build.specification.spec["variables"],
                        )

        utils.symlink(conda_prefix, environment_prefix)

        # modify permissions, uid, gid if they do not match
        stat_info = os.stat(conda_prefix)
        permissions = conda_store.default_permissions
        uid = conda_store.default_uid
        gid = conda_store.default_gid

        if permissions is not None and oct(stat.S_IMODE(stat_info.st_mode))[-3:] != str(
            permissions
        ):
            conda_store.log.info(
                f"modifying permissions of {conda_prefix} to permissions={permissions}"
            )
            with utils.timer(conda_store.log, f"chmod of {conda_prefix}"):
                utils.chmod(conda_prefix, permissions)

        if (
            uid is not None
            and gid is not None
            and (str(uid) != str(stat_info.st_uid) or str(gid) != str(stat_info.st_gid))
        ):
            conda_store.log.info(
                f"modifying permissions of {conda_prefix} to uid={uid} and gid={gid}"
            )
            with utils.timer(conda_store.log, f"chown of {conda_prefix}"):
                utils.chown(conda_prefix, uid, gid)

        packages = conda.conda_prefix_packages(conda_prefix)
        build.size = utils.disk_usage(conda_prefix)

        set_build_completed(conda_store, build, output.encode("utf-8"), packages)
    except subprocess.CalledProcessError as e:
        conda_store.log.exception(e)
        set_build_failed(conda_store, build, e.output.encode("utf-8"))
        raise e
    except Exception as e:
        conda_store.log.exception(e)
        set_build_failed(conda_store, build, traceback.format_exc().encode("utf-8"))
        raise e


def solve_conda_environment(conda_store, solve):
    from conda_store_server.conda import conda_lock

    try:
        solve.started_on = datetime.datetime.utcnow()
        conda_store.db.commit()

        specification = schema.CondaSpecification.parse_obj(solve.specification.spec)
        packages = conda_lock(specification, conda_store.conda_command)

        for package in packages["conda"]:
            channel = package["channel_id"]
            if channel == "https://conda.anaconda.org/pypi":
                # ignore pypi package for now
                continue

            channel_orm = api.get_conda_channel(conda_store.db, channel)
            if channel_orm is None:
                # Empty list for conda_allowed_channels allows any channel ( PR #358 )
                if len(conda_store.conda_allowed_channels) == 0:
                    channel_orm = api.create_conda_channel(conda_store.db, channel)
                    conda_store.db.commit()
                else:
                    raise ValueError(
                        f"channel url={channel} not recognized in conda-store channel database"
                    )
            package["channel_id"] = channel_orm.id

            _package = (
                conda_store.db.query(orm.CondaPackage)
                .filter(orm.CondaPackage.md5 == package["md5"])
                .filter(orm.CondaPackage.channel_id == package["channel_id"])
                .first()
            )

            if _package is None:
                _package = orm.CondaPackage(**package)
                conda_store.db.add(_package)
            solve.packages.append(_package)
        solve.ended_on = datetime.datetime.utcnow()
        conda_store.db.commit()
    except Exception as e:
        print("Task failed!!!!!!!!!!!", str(e))
        raise e


def build_conda_env_export(conda_store, build):
    conda_prefix = build.build_path(conda_store)

    output = subprocess.check_output(
        [conda_store.conda_command, "env", "export", "-p", conda_prefix]
    )

    parsed = yaml.safe_load(output)
    if "dependencies" not in parsed:
        raise ValueError(f"conda env export` did not produce valid YAML:\n{output}")

    conda_store.storage.set(
        conda_store.db,
        build.id,
        build.conda_env_export_key,
        output,
        content_type="text/yaml",
        artifact_type=schema.BuildArtifactType.YAML,
    )


def build_conda_pack(conda_store, build):
    conda_prefix = build.build_path(conda_store)

    conda_store.log.info(f"packaging archive of conda environment={conda_prefix}")
    with tempfile.TemporaryDirectory() as tmpdir:
        output_filename = os.path.join(tmpdir, "environment.tar.gz")
        conda.conda_pack(prefix=conda_prefix, output=output_filename)
        conda_store.storage.fset(
            conda_store.db,
            build.id,
            build.conda_pack_key,
            output_filename,
            content_type="application/gzip",
            artifact_type=schema.BuildArtifactType.CONDA_PACK,
        )


def build_conda_docker(conda_store, build):
    from conda_docker.conda import (
        build_docker_environment_image,
        find_user_conda,
        conda_info,
        precs_from_environment_prefix,
        fetch_precs,
    )

    conda_prefix = build.build_path(conda_store)

    conda_store.log.info(f"creating docker archive of conda environment={conda_prefix}")

    user_conda = find_user_conda()
    info = conda_info(user_conda)
    download_dir = info["pkgs_dirs"][0]
    precs = precs_from_environment_prefix(conda_prefix, download_dir, user_conda)
    records = fetch_precs(download_dir, precs)
    base_image = conda_store.container_registry.pull_image(
        utils.callable_or_value(conda_store.default_docker_base_image, build)
    )
    image = build_docker_environment_image(
        base_image=base_image,
        output_image=f"{build.specification.name}:{build.build_key}",
        records=records,
        default_prefix=info["env_vars"]["CONDA_ROOT"],
        download_dir=download_dir,
        user_conda=user_conda,
        channels_remap=info.get("channels_remap", []),
        layering_strategy="layered",
    )

    if schema.BuildArtifactType.DOCKER_MANIFEST in conda_store.build_artifacts:
        conda_store.container_registry.store_image(conda_store, build, image)

    if schema.BuildArtifactType.CONTAINER_REGISTRY in conda_store.build_artifacts:
        conda_store.container_registry.push_image(conda_store, build, image)



"""Adding CONTAINER_REGISTRY value to enum

Revision ID: 5ad723de2abd
Revises: 8d63a091aff8
Create Date: 2022-08-05 22:14:34.110642

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "5ad723de2abd"
down_revision = "8d63a091aff8"
branch_labels = None
depends_on = None


def upgrade():
    old_type = sa.Enum(
        "DIRECTORY",
        "LOCKFILE",
        "LOGS",
        "YAML",
        "CONDA_PACK",
        "DOCKER_BLOB",
        "DOCKER_MANIFEST",
        name="buildartifacttype",
    )

    new_type = sa.Enum(
        "DIRECTORY",
        "LOCKFILE",
        "LOGS",
        "YAML",
        "CONDA_PACK",
        "DOCKER_BLOB",
        "DOCKER_MANIFEST",
        "CONTAINER_REGISTRY",
        name="buildartifacttype",
    )

    with op.batch_alter_table("build_artifact") as batch_op:
        batch_op.alter_column("artifact_type", type_=new_type, existing_type=old_type)


def downgrade():
    # harmless to keep extra enum around
    pass



import logging
import os
import sys

from traitlets import Unicode, Integer, List, validate
from traitlets.config import Application, catch_config_error

from conda_store_server.app import CondaStore


class CondaStoreWorker(Application):
    aliases = {
        "config": "CondaStoreWorker.config_file",
    }

    flags = {
        "standalone": (
            {
                "CondaStoreServer": {
                    "standalone": True,
                }
            },
            "Run conda-store-server in standalone mode with celery worker as a subprocess of webserver",
        ),
    }

    log_level = Integer(
        logging.INFO,
        help="log level to use",
        config=True,
    )

    watch_paths = List(
        [], help="list of paths to watch for environment changes", config=True
    )

    concurrency = Integer(
        None,
        help="Number of worker threads to spawn. Limit the concurrency of conda-builds",
        config=True,
        allow_none=True,
    )

    config_file = Unicode(
        help="config file to load for conda-store",
        config=True,
    )

    @validate("config_file")
    def _validate_config_file(self, proposal):
        if not os.path.isfile(proposal.value):
            print(
                "ERROR: Failed to find specified config file: {}".format(
                    proposal.value
                ),
                file=sys.stderr,
            )
            sys.exit(1)
        return proposal.value

    @catch_config_error
    def initialize(self, *args, **kwargs):
        super().initialize(*args, **kwargs)
        self.load_config_file(self.config_file)

        self.conda_store = CondaStore(parent=self, log=self.log)

        # ensure checks on redis_url
        self.conda_store.redis_url

    def start(self):
        argv = [
            "worker",
            "--loglevel=INFO",
            "--beat",
        ]

        if self.concurrency:
            argv.append(f"--concurrency={self.concurrency}")

        self.conda_store.ensure_directories()
        self.conda_store.celery_app.worker_main(argv)



import hashlib
import gzip
import urllib.parse

from traitlets.config import LoggingConfigurable
from traitlets import Dict, Callable, default
from python_docker.registry import Image, Registry

from conda_store_server import schema, orm, utils


class ContainerRegistry(LoggingConfigurable):
    container_registries = Dict(
        {},
        help="Registries url to upload built container images with callable function to configure registry instance with credentials",
        config=True,
    )

    container_registry_image_name = Callable(
        help="Image name to assign to docker image pushed for particular registry",
        config=True,
    )

    @default("container_registry_image_name")
    def _default_container_registry_image_name(self):
        def _container_registry_image_name(registry: Registry, build: orm.Build):
            return f"{registry.username}/{build.environment.namespace.name}-{build.environment.name}"

        return _container_registry_image_name

    container_registry_image_tag = Callable(
        help="Image name and tag to assign to docker image pushed for particular registry",
        config=True,
    )

    @default("container_registry_image_tag")
    def _default_container_registry_image_tag(self):
        def _container_registry_image_tag(registry: Registry, build: orm.Build):
            return build.key

        return _container_registry_image_tag

    def store_image(self, conda_store, build: orm.Build, image: Image):
        self.log.info("storing container image locally")
        with utils.timer(self.log, "storing container image locally"):
            # https://docs.docker.com/registry/spec/manifest-v2-2/#example-image-manifest
            docker_manifest = schema.DockerManifest.construct()
            docker_config = schema.DockerConfig.construct(
                config=schema.DockerConfigConfig(),
                container_config=schema.DockerConfigConfig(),
                rootfs=schema.DockerConfigRootFS(),
            )

            for layer in image.layers:
                # https://github.com/google/nixery/pull/64#issuecomment-541019077
                # docker manifest expects compressed hash while configuration file
                # expects uncompressed hash -- good luck finding this detail in docs :)
                content_uncompressed_hash = hashlib.sha256(layer.content).hexdigest()
                content_compressed = gzip.compress(layer.content)
                content_compressed_hash = hashlib.sha256(content_compressed).hexdigest()
                conda_store.storage.set(
                    conda_store.db,
                    build.id,
                    build.docker_blob_key(content_compressed_hash),
                    content_compressed,
                    content_type="application/gzip",
                    artifact_type=schema.BuildArtifactType.DOCKER_BLOB,
                )

                docker_layer = schema.DockerManifestLayer(
                    size=len(content_compressed),
                    digest=f"sha256:{content_compressed_hash}",
                )
                docker_manifest.layers.append(docker_layer)

                docker_config_history = schema.DockerConfigHistory()
                docker_config.history.append(docker_config_history)

                docker_config.rootfs.diff_ids.append(
                    f"sha256:{content_uncompressed_hash}"
                )

            docker_config_content = docker_config.json().encode("utf-8")
            docker_config_hash = hashlib.sha256(docker_config_content).hexdigest()
            docker_manifest.config = schema.DockerManifestConfig(
                size=len(docker_config_content), digest=f"sha256:{docker_config_hash}"
            )
            docker_manifest_content = docker_manifest.json().encode("utf-8")
            docker_manifest_hash = hashlib.sha256(docker_manifest_content).hexdigest()

            conda_store.storage.set(
                conda_store.db,
                build.id,
                build.docker_blob_key(docker_config_hash),
                docker_config_content,
                content_type="application/vnd.docker.container.image.v1+json",
                artifact_type=schema.BuildArtifactType.DOCKER_BLOB,
            )

            # docker likes to have a sha256 key version of the manifest this
            # is sort of hack to avoid having to figure out which sha256
            # refers to which manifest.
            conda_store.storage.set(
                conda_store.db,
                build.id,
                f"docker/manifest/sha256:{docker_manifest_hash}",
                docker_manifest_content,
                content_type="application/vnd.docker.distribution.manifest.v2+json",
                artifact_type=schema.BuildArtifactType.DOCKER_BLOB,
            )

            conda_store.storage.set(
                conda_store.db,
                build.id,
                build.docker_manifest_key,
                docker_manifest_content,
                content_type="application/vnd.docker.distribution.manifest.v2+json",
                artifact_type=schema.BuildArtifactType.DOCKER_MANIFEST,
            )

            conda_store.log.info(
                f"built docker image: {image.name}:{image.tag} layers={len(image.layers)}"
            )

    @staticmethod
    def parse_image_uri(image_name: str):
        """Must be in fully specified format [<scheme>://]<registry_url>/<image_name>:<tag_name>"""
        if not image_name.startswith("http"):
            image_name = f"https://{image_name}"

        parsed_url = urllib.parse.urlparse(image_name)
        registry_url = f"{parsed_url.scheme}://{parsed_url.netloc}"
        image_name, tag_name = parsed_url.path.split(":", 1)
        image_name = image_name[1:]  # remove beginning "/"
        return registry_url, image_name, tag_name

    def pull_image(self, image_name: str) -> Image:
        """Must be in fully specified format [<scheme>://]<registry_url>/<image_name>:<tag_name>

        Docker is the only weird registry where you must use:
          - `https://registry-1.docker.io`
        """
        registry_url, name, tag = self.parse_image_uri(image_name)

        for url in self.container_registries:
            if registry_url in url:
                registry = self.container_registries[registry_url](url)
                break
        else:
            self.log.warning(
                f"registry {registry_url} not configured using registry without authentication"
            )
            registry = Registry(hostname=registry_url)

        return registry.pull_image(name, tag)

    def push_image(self, conda_store, build, image: Image):
        for registry_url, configure_registry in self.container_registries.items():
            self.log.info(f"beginning upload of image to registry {registry_url}")
            with utils.timer(self.log, f"uploading image to registry {registry_url}"):
                registry = configure_registry(registry_url)
                image.name = self.container_registry_image_name(registry, build)
                image.tag = self.container_registry_image_tag(registry, build)
                registry.push_image(image)

                registry_build_artifact = orm.BuildArtifact(
                    build_id=build.id,
                    artifact_type=schema.BuildArtifactType.CONTAINER_REGISTRY,
                    key=f"{registry_url}/{image.name}:{image.tag}",
                )
                conda_store.db.add(registry_build_artifact)
                conda_store.db.commit()

    def delete_image(self, image_name: str):
        registry_url, name, tag = self.parse_image_uri(image_name)

        for url in self.container_registries:
            if registry_url in url:
                registry = self.container_registries[registry_url](url)
                break
        else:
            self.log.warning(
                f"registry {registry_url} not configured using registry without authentication"
            )
            registry = Registry(hostname=registry_url)

        self.log.info(f"deleting container image {image_name}")
        registry.delete_image(name, tag)
